@@ -1,10 +1,10 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 
 #
-# Copyright (C) 2019 Jerry Hoogenboom
+# Copyright (C) 2021 Jerry Hoogenboom
 #
-# This file is part of FDSTools, data analysis tools for Next
-# Generation Sequencing of forensic DNA markers.
+# This file is part of FDSTools, data analysis tools for Massively
+# Parallel Sequencing of forensic DNA markers.
 #
 # FDSTools is free software: you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -47,19 +47,30 @@ tools in FDSTools.  Please refer to the tool-specific help for a full
 description of each tool.  Type 'fdstools -h TOOL' to get help with the
 given TOOL.
 """
-import pkgutil, sys, os, tempfile, re, argparse
-#import threading, subprocess  # Only imported when running this tool.
+import os
+import pkgutil
+import re
+import subprocess
+import sys
+import tempfile
+import threading
 
 import fdstools.tools
 
-from ..lib import split_quoted_string, DEF_TAG_EXPR, DEF_TAG_FORMAT, get_tag, \
-                  regex_arg, INI_COMMENT, glob_path, print_db
-from ConfigParser import RawConfigParser, NoSectionError, NoOptionError
+from configparser import RawConfigParser, NoSectionError, NoOptionError
+from functools import reduce
+
+from ..lib.cli import DEF_TAG_EXPR, DEF_TAG_FORMAT, get_tag, regex_arg, glob_path
+from ..lib.io import print_db
+from ..lib.library import INI_COMMENT
 
 
-__version__ = "1.0.4"
+__version__ = "1.1.1"
 
 
+# Pattern to split a quoted string.
+PAT_SPLIT_QUOTED = re.compile(r""""((?:\\"|[^"])*)"|'((?:\\'|[^'])*)'|(\S+)""")
+
 # Pattern that matches a long argparse argument name.
 PAT_ARGNAME = re.compile("^(?:--)?([a-z0-9]+(?:-[a-z0-9]+)*)$")
 
@@ -73,12 +84,10 @@ PACKAGE_PREFIX = fdstools.tools.__name__ + "."
 NAME = __name__[len(PACKAGE_PREFIX):]
 
 # Argument names that will not be written in generated INI files.
-HIDDEN_ARGS = ("infile", "infile2", "infiles", "outfile", "outfiles",
-    "raw-outfile", "library", "library2", "allelelist", "report", "tag-expr",
-    "tag-format", "sequence-format", "type", "title", "stuttermodel", "seqs",
-    "profiles", "column-name", "stuttermark-column", "annotation-column",
-    "marker-column", "allele-column", "output-column", "reverse-complement",
-    "limit-reads", "drop-samples")
+HIDDEN_ARGS = ("infile", "infile2", "infiles", "outfile", "outfiles", "raw-outfile",
+    "library", "library2", "allelelist", "report", "tag-expr", "tag-format", "uncall-alleles",
+    "sequence-format", "type", "title", "stuttermodel", "seqs", "profiles", "combine-strands",
+    "annotation-column", "marker-column", "allele-column", "output-column", "reverse-complement")
 
 # Tools that are run for each type of analysis pipeline.
 ANALYSIS_TOOLS = {
@@ -98,15 +107,10 @@ ANALYSIS_ARGS = {
         ("analysis", "in-library", "in-sample-raw", "tag-expr", "tag-format"),
     "reference-database":
         ("analysis", "in-library", "in-samples", "in-allelelist", "prefix",
-        "tag-expr", "tag-format"),
+        "tag-expr", "tag-format", "combine-strands"),
     "case-sample":
         ("analysis", "in-library", "in-sample-raw", "in-stuttermodel",
-        "in-bgprofiles", "store-predictions", "tag-expr", "tag-format")
-}
-
-# Pipeline default argument values overriding the tool's own defaults.
-DEFAULT_OVERRIDE = {
-    "tssv": {"aggregate-filtered": True}
+        "in-bgprofiles", "store-predictions", "tag-expr", "tag-format", "combine-strands")
 }
 
 
@@ -148,14 +152,12 @@ class ArgumentCollector:
     def add_argument(self, tool, *args, **kwargs):
         # Determine argument name.
         if "dest" in kwargs:
-            name = kwargs["dest"]
+            name = kwargs["dest"].replace("_", "-")
         else:
             for arg in args:
                 match = PAT_ARGNAME.match(arg)
                 if match is not None:
                     name = match.group(1)
-                    if arg.startswith("--") and "nargs" not in kwargs:
-                        kwargs["nargs"] = "?"
                     break
             else:
                 name = args[0]
@@ -163,19 +165,26 @@ class ArgumentCollector:
             kwargs["optname"] = args[0]
 
         # Replace special default values.
-        if "default" in kwargs:
-            if kwargs["default"] is None or kwargs["default"] is "":
+        if "optname" in kwargs and "nargs" in kwargs and kwargs["nargs"] == "?":
+            # Value can be False (option unspecified),
+            # True (option specified),
+            # or an actual value (option specified with argument).
+            kwargs["default"] = False
+        elif "default" in kwargs:
+            if kwargs["default"] is None or kwargs["default"] == "":
                 del kwargs["default"]
             elif kwargs["default"] in (sys.stdin, sys.stdout):
                 kwargs["default"] = "-"
             elif kwargs["default"] is sys.stderr:
                 kwargs["default"] = "/dev/stderr"
-            elif kwargs["default"] is sys.maxint:
+            elif kwargs["default"] is sys.maxsize:
                 kwargs["default"] = "<MAX>"
             elif isinstance(kwargs["default"], list):
                 kwargs["default"] = " ".join(kwargs["default"])
-        if "action" in kwargs and kwargs["action"] == "store_true":
+        elif "action" in kwargs and kwargs["action"] == "store_true":
             kwargs["default"] = False
+        elif "action" in kwargs and kwargs["action"] == "store_false":
+            kwargs["default"] = True
         self.arglists[tool].append((name, kwargs))
     #add_argument
 
@@ -185,6 +194,14 @@ class ArgumentCollector:
 #ArgumentCollector
 
 
+def split_quoted_string(text):
+    return ["".join((
+        y[0].replace("\\\"", "\""),
+        y[1].replace("\\'", "'"),
+        y[2])) for y in PAT_SPLIT_QUOTED.findall(text)]
+#split_quoted_string
+
+
 def format_help(text, format_args={}):
     text = PAT_ARGNAME_IN_DOC.sub("'\\1'",
         text.replace(
@@ -200,8 +217,12 @@ def get_tools(hide_pipeline):
             path=fdstools.tools.__path__):
         if hide_pipeline and name == NAME:
             continue
-        tools[name] = importer.find_module(PACKAGE_PREFIX + name).load_module(
-            PACKAGE_PREFIX + name)
+        try:
+            tools[name] = importer.find_module(PACKAGE_PREFIX + name).load_module(
+                PACKAGE_PREFIX + name)
+        except Exception as error:
+            sys.stderr.write("FDSTools failed to load '%s': %s\n" % (name, error))
+            continue
     return tools
 #get_tools
 
@@ -236,6 +257,8 @@ def parse_bool_arg(section, option, value):
 def get_argv(toolname, arg_defs, config):
     """Get argument list for the given tool."""
     arglist = (["fdstools", toolname], [])
+    if debug:
+        arglist[0].append("-d")
     for arg in arg_defs[toolname]:
         try:
             value = config.get(toolname, arg[0]).strip()
@@ -250,23 +273,48 @@ def get_argv(toolname, arg_defs, config):
             if parse_bool_arg(toolname, arg[0], value):
                 arglist[0].append(arg[1]["optname"])
             continue
+        if "action" in arg[1] and arg[1]["action"] == "store_false":
+            if not parse_bool_arg(toolname, arg[0], value):
+                arglist[0].append(arg[1]["optname"])
+            continue
         if "default" in arg[1] and arg[1]["default"] == value:
             if "optname" in arg[1]:
                 continue
-            if "default" in arg[1] and arg[1]["default"] is "<MAX>":
-                value = str(sys.maxint)
+            if value == "<MAX>":
+                value = str(sys.maxsize)
         if "nargs" in arg[1] and arg[1]["nargs"] not in (1, "?"):
             value = split_quoted_string(value)
+        elif "optname" in arg[1] and "nargs" in arg[1] and arg[1]["nargs"] == "?":
+            # Value can be False (option unspecified),
+            # True (option specified),
+            # or an actual value (option specified with argument).
+            try:
+                if value in ("0", "1"):
+                    # Check if 0 or 1 is a valid argument value.
+                    arg[1]["type"](value)
+                    value = [value]  # Specify with given argument.
+                else:
+                    raise ValueError  # Proceed to next check.
+            except:
+                try:
+                    # Try to parse the value as a boolean (no argument).
+                    value = parse_bool_arg(toolname, arg[0], value)
+                    if not value:
+                        continue  # Value False, don't specify option.
+                    value = None  # Value True, specify without arg.
+                except ValueError:
+                    value = [value]  # Specify with given argument.
         else:
             value = [value]
         if "optname" in arg[1]:
-            if len(value) == 1 and value[0].startswith("-"):
+            if value is not None and len(value) == 1 and value[0].startswith("-"):
                 # Fix Argparse's "looks like an option" oddity.
                 arglist[0].append("%s=%s" % (arg[1]["optname"], value[0]))
             else:
                 arglist[0].append(arg[1]["optname"])
-                arglist[0].extend(value)
-        else:
+                if value is not None:
+                    arglist[0].extend(value)
+        elif value is not None:
             arglist[1].extend(value)
     print_db(("%r" % (arglist[0]+arglist[1]))[:200], debug)
     return arglist[0] + arglist[1]
@@ -422,6 +470,7 @@ def run_ref_database_analysis(arg_defs, config):
     if allelefile is None:
         allelefile =  prefix + "allelelist.txt"
         run_allelefinder = True
+    combine_strands = ini_try_get_option(config, NAME, "combine-strands", "")
 
     # Overwrite any explicit I/O configuration.
     if run_allelefinder:
@@ -443,6 +492,7 @@ def run_ref_database_analysis(arg_defs, config):
     config.set("bgestimate", "library", in_library)
     config.set("bgestimate", "tag-expr", tag_expr)
     config.set("bgestimate", "tag-format", tag_format)
+    config.set("bgestimate", "combine-strands", combine_strands)
     if not config.has_section("stuttermodel"):
         config.add_section("stuttermodel")
     config.set("stuttermodel", "allelelist", allelefile)
@@ -452,6 +502,7 @@ def run_ref_database_analysis(arg_defs, config):
     config.set("stuttermodel", "tag-expr", tag_expr)
     config.set("stuttermodel", "tag-format", tag_format)
     config.set("stuttermodel", "raw-outfile", prefix + "stuttermodel-raw.txt")
+    config.set("stuttermodel", "combine-strands", combine_strands)
     if not config.has_section("bghomraw"):
         config.add_section("bghomraw")
     config.set("bghomraw", "allelelist", allelefile)
@@ -460,6 +511,7 @@ def run_ref_database_analysis(arg_defs, config):
     config.set("bghomraw", "library", in_library)
     config.set("bghomraw", "tag-expr", tag_expr)
     config.set("bghomraw", "tag-format", tag_format)
+    config.set("bghomraw", "combine-strands", combine_strands)
     if not config.has_option("bghomraw", "sequence-format"):
         config.set("bghomraw", "sequence-format", "allelename")
     if not config.has_section("vis"):
@@ -566,6 +618,7 @@ def run_case_sample_analysis(arg_defs, config):
             config, NAME, "store-predictions", "False"))
     out_bgpredict = tag + "-bgpredict.txt" if store_predictions else "-"
     out_bgmerge = tag + "-bgmerge.txt" if store_predictions else "-"
+    combine_strands = ini_try_get_option(config, NAME, "combine-strands", "")
 
 
     # Overwrite any explicit I/O configuration.
@@ -582,6 +635,7 @@ def run_case_sample_analysis(arg_defs, config):
         config.set("bgpredict", "seqs", "-")
         config.set("bgpredict", "outfile", out_bgpredict)
         config.set("bgpredict", "library", in_library)
+        config.set("bgpredict", "combine-strands", combine_strands)
         if in_bgprofiles is not None:
             if not config.has_section("bgmerge"):
                 config.add_section("bgmerge")
@@ -603,6 +657,7 @@ def run_case_sample_analysis(arg_defs, config):
         config.set("bgcorrect", "outfile", "-")
         config.remove_option("bgcorrect", "outfiles")
         config.set("bgcorrect", "library", in_library)
+        config.set("bgcorrect", "combine-strands", combine_strands)
     if not config.has_section("seqconvert"):
         config.add_section("seqconvert")
     if not config.has_option("seqconvert", "sequence-format"):
@@ -640,7 +695,7 @@ def run_case_sample_analysis(arg_defs, config):
             if config.get("bgcorrect", "profiles") == "-":
                 # Set up FIFO if BGCorrect gets two pipes in!
                 tmpfile[0] = tempfile.mkdtemp()
-                tmpfile[1] = os.path.join(tmpfile[0], 'fifo.tmp')
+                tmpfile[1] = os.path.join(tmpfile[0], "fifo.tmp")
                 try:
                     os.mkfifo(tmpfile[1])
                     tmpfile[2] = True  # Is a FIFO.
@@ -753,9 +808,9 @@ def run_case_sample_analysis(arg_defs, config):
 
 
 def add_ini_arg(argument, value, config):
-    if value is None:
+    if value is None or value == False:
         return
-    if type(value) is list:
+    if isinstance(value, list):
         value = " ".join(
             "'" + x.replace("'", "\'") + "'" for x in value)
     else:
@@ -787,7 +842,7 @@ def run_ini(config):
     analysis = ini_require_option(config, NAME, "analysis")
     if analysis not in analysis_functions:
         raise ValueError("Unknown analysis type '%s', specify one of: %s" %
-            (analysis, ", ".join(analysis_functions.keys())))
+            (analysis, ", ".join(analysis_functions)))
     analysis_functions[analysis](get_arguments(get_tools(True)), config)
 #run_ini
 
@@ -797,11 +852,10 @@ def write_ini(args):
     arglists = get_arguments(tools)
     for name in ANALYSIS_TOOLS[args.analysis]:
         if name == NAME:
-            fmt = "%%-%is" % reduce(max, map(len,
-                ANALYSIS_ARGS[args.analysis]), 0)
+            fmt = "%%-%is" % max(map(len, ANALYSIS_ARGS[args.analysis] or [""]))
         else:
-            fmt = "%%-%is" % reduce(max, map(len,
-                (x[0] for x in arglists[name] if x[0] not in HIDDEN_ARGS)), 0)
+            fmt = "%%-%is" % reduce(max,
+                (len(x[0]) for x in arglists[name] if x[0] not in HIDDEN_ARGS), 0)
         if fmt == "%-0s":
             continue  # No arguments.
         args.config.write("[%s]\n%s\n\n" % (name,
@@ -814,8 +868,6 @@ def write_ini(args):
                     arg[1]["default"] = getattr(args, arg[0].replace("-","_"))
             elif arg[0] in HIDDEN_ARGS:
                 continue  # Argument is being overridden by this tool.
-            elif name in DEFAULT_OVERRIDE and arg[0] in DEFAULT_OVERRIDE[name]:
-                arg[1]["default"] = DEFAULT_OVERRIDE[name][arg[0]]
             if "help" in arg[1]:
                 args.config.write("%s\n" % format_help(arg[1]["help"], arg[1]))
             if "default" in arg[1]:
@@ -845,12 +897,12 @@ def add_arguments(parser):
         "these options are used to extract sample tags (names) from their "
         "file names; for details about REGEX syntax and capturing groups, "
         "check https://docs.python.org/howto/regex")
-    group.add_argument('-e', '--tag-expr', metavar="REGEX",
+    group.add_argument("-e", "--tag-expr", metavar="REGEX",
         help="regular expression that captures (using one or more "
              "capturing groups) the sample tags from the file names; by "
              "default, the entire file name except for its extension (if "
              "any) is captured")
-    group.add_argument('-f', '--tag-format', metavar="EXPR",
+    group.add_argument("-f", "--tag-format", metavar="EXPR",
         help="format of the sample tags produced; a capturing group "
              "reference like '\\n' refers to the n-th capturing group in "
              "the regular expression specified with -e/--tag-expr (the "
@@ -879,7 +931,7 @@ def add_arguments(parser):
     group.add_argument("-S", "--in-samples", metavar="SAMPLE", nargs="+",
         help="[ref-database] file names of reference sample data files "
              "('.csv' output files of the 'reference-sample' analysis)")
-    group.add_argument('-A', '--in-allelelist', metavar="ALLELEFILE",
+    group.add_argument("-A", "--in-allelelist", metavar="ALLELEFILE",
         help="[ref-database] file containing a list of the true alleles of "
              "each sample; if not given, Allelefinder will be run as part of "
              "the pipeline to create this file; it is ESSENTIAL that you "
@@ -887,21 +939,13 @@ def add_arguments(parser):
     group.add_argument("-P", "--prefix",
         help="[ref-database] if specified, all output file names are prefixed "
              "with this value")
+    group.add_argument("-C", "--combine-strands", action="store_true",
+        help="[ref-database, case-sample] if specified, noise analysis will be performed on the "
+             "total number of reads, instead of separately for either strand")
 #add_arguments
 
 
 def run(args):
-    # Import threading and subprocess modules.
-    global threading, subprocess
-    import threading
-    if os.name == 'posix' and sys.version_info[0] < 3:
-        try:
-            import subprocess32 as subprocess
-        except ImportError:
-            import subprocess
-    else:
-        import subprocess
-
     # Export debug flag to the global scope.
     global debug
     debug = args.debug
@@ -910,7 +954,7 @@ def run(args):
         raise ValueError("The pipeline configuration file cannot be named '-'")
     if os.path.exists(args.config):
         # Run analysis pipeline using configuration file.
-        args.config = open(args.config, "r")
+        args.config = open(args.config, "rt", encoding="UTF-8")
         run_ini(get_config(args))
         args.config.close()
     elif args.analysis is not None:
@@ -918,7 +962,7 @@ def run(args):
         print("Configuration file '%s' does not exist, creating default "
               "configuration file for %s analysis..." %
               (args.config, args.analysis))
-        args.config = open(args.config, "w")
+        args.config = open(args.config, "wt", encoding="UTF-8")
         write_ini(args)
         args.config.close()
     else:
