@@ -1,10 +1,10 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 
 #
-# Copyright (C) 2019 Jerry Hoogenboom
+# Copyright (C) 2022 Jerry Hoogenboom
 #
-# This file is part of FDSTools, data analysis tools for Next
-# Generation Sequencing of forensic DNA markers.
+# This file is part of FDSTools, data analysis tools for Massively
+# Parallel Sequencing of forensic DNA markers.
 #
 # FDSTools is free software: you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -59,27 +59,32 @@ References:
 [1] https://github.com/Jerrythafast/FLASH-lowercase-overhang
 [2] https://github.com/jfjlaros/tssv
 """
-import sys
+import gzip
+import io
+import itertools
 import math
-import itertools as it
 import os
+import sys
 
-from multiprocessing.queues import SimpleQueue
-from multiprocessing import Process
-from threading import Thread, Lock
 from errno import EPIPE
+from multiprocessing import Process, SimpleQueue
+from threading import Thread, Lock
 
+from ..lib.cli import add_sequence_format_args, add_input_output_args, pos_int_arg,\
+                      get_input_output_files
+from ..lib.io import try_write_pipe
+from ..lib.seq import PAT_SEQ_RAW, reverse_complement, ensure_sequence_format
+from ..lib.sg_align import align
 
-from ..lib import pos_int_arg, add_input_output_args, get_input_output_files,\
-                  add_sequence_format_args, reverse_complement, PAT_SEQ_RAW,\
-                  ensure_sequence_format
-from ..sg_align import align
-
-__version__ = "2.0.0"
+__version__ = "2.1.1"
 
 
 # Default values for parameters are specified below.
 
+# Default flanking sequence length.
+# This value can be overridden by the -L command line option.
+_DEF_FLANK_LENGTH = 16
+
 # Default maximum number of mismatches per nucleotide in the flanking
 # sequences to allow.
 # This value can be overridden by the -m command line option.
@@ -92,12 +97,15 @@ _DEF_INDEL_SCORE = 2
 
 # Default minimum number of reads to consider.
 # This value can be overridden by the -a command line option.
-_DEF_MINIMUM = 1
+_DEF_MINIMUM = 2
+
+IUPAC_BITS = str.maketrans("ACGTUMRSWYKVHDBN", "ABDHHCEFIJLGKMNO")
 
 
 
 class TSSV:
-    def __init__(self, library, threshold, indel_score, dirname, workers, deduplicate, infile):
+    def __init__(self, library, flank_length, threshold, indel_score, dirname, workers,
+                 deduplicate, infile):
         # User inputs.
         self.library = library
         self.indel_score = indel_score
@@ -106,11 +114,33 @@ class TSSV:
         self.lock = Lock()
 
         # Convert library.
-        self.tssv_library = {marker: (
-            (flanks[0], reverse_complement(flanks[1])), (
-                int(math.ceil(len(flanks[0]) * threshold)),
-                int(math.ceil(len(flanks[1]) * threshold))))
-            for marker, flanks in library["flanks"].items()}
+        refseq_store = library.get_structure_store().get_refseq_store()
+        self.has_iupac = False
+        self.tssv_library = {}
+        for marker, reported_range in library.get_ranges().items():
+            flanks = list(reported_range.get_option("flanks", (flank_length, flank_length)))
+            if not all(flanks):
+                raise ValueError("Missing flanking sequence for marker %s" % marker)
+            chromosome, start, *_, end = reported_range.location
+            if isinstance(flanks[0], int):
+                flanks[0] = refseq_store.get_refseq(chromosome, start - flanks[0], start)
+            if isinstance(flanks[1], int):
+                flanks[1] = refseq_store.get_refseq(chromosome, end + 1, end + 1 + flanks[1])
+            flanks[1] = reverse_complement(flanks[1])
+            translate_flank = [0 if PAT_SEQ_RAW.match(flank) else 1 for flank in flanks]
+            if any(translate_flank):
+                self.has_iupac = True
+            pair = tuple((flank.translate(IUPAC_BITS) if trans else flank, trans)
+                for flank, trans in zip(flanks, translate_flank))
+            self.tssv_library[marker] = (
+                pair, (
+                    math.ceil(len(flanks[0]) * threshold if threshold < 1 else threshold),
+                    math.ceil(len(flanks[1]) * threshold if threshold < 1 else threshold)),
+                "".join(refseq_store.get_refseq(chromosome, start, end + 1) for start, end in
+                    zip(reported_range.location[1::2], reported_range.location[2::2])))
+
+        if not self.tssv_library:
+            raise ValueError("No markers were defined in the given library file")
 
         # Open input file.
         file_format, self.input = init_sequence_file_read(infile)
@@ -151,14 +181,13 @@ class TSSV:
 
 
     def cache_results(self, seq, results):
-        self.lock.acquire()
+        """NOTE: In a concurrent context, caller must hold self.lock!"""
         for record in self.cache[seq][1]:
             self.process_results(tuple([record[0], seq]) + record[1:], results)
         if self.deduplicate:
             self.cache[seq] = (True, results)
         else:
             del self.cache[seq]
-        self.lock.release()
     #cache_results
 
 
@@ -167,39 +196,45 @@ class TSSV:
         recognised = 0
         for marker, matches, seq1, seq2 in results:
             recognised |= matches
-            self.counters[marker]["fLeft"] += matches & 1
-            self.counters[marker]["fRight"] += matches >> 1 & 1
-            self.counters[marker]["rLeft"] += matches >> 2 & 1
-            self.counters[marker]["rRight"] += matches >> 3 & 1
+            sequences = self.sequences[marker]
+            counters = self.counters[marker]
+            if matches & 0b0001:
+                counters["fLeft"] += 1
+            if matches & 0b0010:
+                counters["fRight"] += 1
+            if matches & 0b0100:
+                counters["rLeft"] += 1
+            if matches & 0b1000:
+                counters["rRight"] += 1
 
             # Search in the forward strand.
             if seq1 is not None:
-                self.counters[marker]["fPaired"] += 1
-                if seq1 not in self.sequences[marker]:
-                    self.sequences[marker][seq1] = [1, 0]
-                else:
-                    self.sequences[marker][seq1][0] += 1
+                counters["fPaired"] += 1
+                try:
+                    sequences[seq1][0] += 1
+                except KeyError:
+                    sequences[seq1] = [1, 0]
                 if self.outfiles:
                     write_sequence_record(self.outfiles["markers"][marker]["paired"], record)
             elif self.outfiles:
-                if matches & 1:
+                if matches & 0b0001:
                     write_sequence_record(self.outfiles["markers"][marker]["noend"], record)
-                if matches & 2:
+                if matches & 0b0010:
                     write_sequence_record(self.outfiles["markers"][marker]["nostart"], record)
 
             # Search in the reverse strand.
             if seq2 is not None:
-                self.counters[marker]["rPaired"] += 1
-                if seq2 not in self.sequences[marker]:
-                    self.sequences[marker][seq2] = [0, 1]
-                else:
-                    self.sequences[marker][seq2][1] += 1
+                counters["rPaired"] += 1
+                try:
+                    sequences[seq2][1] += 1
+                except KeyError:
+                    sequences[seq2] = [0, 1]
                 if self.outfiles:
                     write_sequence_record(self.outfiles["markers"][marker]["paired"], record)
             elif self.outfiles:
-                if matches & 4:
+                if matches & 0b0100:
                     write_sequence_record(self.outfiles["markers"][marker]["noend"], record)
-                if matches & 8:
+                if matches & 0b1000:
                     write_sequence_record(self.outfiles["markers"][marker]["nostart"], record)
 
         if not recognised:
@@ -212,18 +247,27 @@ class TSSV:
     def process_file(self):
         if self.workers == 1:
             for seq in self.dedup_reads():
-                self.cache_results(seq, process_sequence(self.tssv_library, self.indel_score, seq))
+                self.cache_results(seq, process_sequence(
+                    self.tssv_library, self.indel_score, self.has_iupac, seq))
         else:
             # Start worker processes.  The work is divided into tasks that
             # require about 1 million alignments each.
             done_queue = SimpleQueue()
-            chunksize = int(1000000/(4*len(self.tssv_library))) or 1
+            chunksize = 1000000 // (4 * len(self.tssv_library)) or 1
             thread = Thread(target=feeder, args=(self.dedup_reads(), self.tssv_library,
-                self.indel_score, self.workers, chunksize, done_queue))
+                self.indel_score, self.has_iupac, self.workers, chunksize, done_queue))
             thread.daemon = True
             thread.start()
-            for seq, results in it.chain.from_iterable(iter(done_queue.get, None)):
-                self.cache_results(seq, results)
+
+            # Process the results as they come in.
+            # Below is speed-optimized to manage as many workers as possible.
+            acquire_lock = self.lock.acquire
+            release_lock = self.lock.release
+            cache_results = self.cache_results
+            for seq, results in itertools.chain.from_iterable(iter(done_queue.get, None)):
+                acquire_lock()
+                cache_results(seq, results)
+                release_lock()
             thread.join()
 
         # Count number of unique sequences per marker.
@@ -236,9 +280,9 @@ class TSSV:
         # Aggregate the sequences that we are about to filter out.
         if aggregate_filtered:
             aggregates = {}
-            for marker in self.sequences:
-                expected_length = self.library.get("expected_length", {}).get(marker)
-                for sequence, (forward, reverse) in self.sequences[marker].items():
+            for marker, sequences in self.sequences.items():
+                expected_length = self.library.get_range(marker).get_option("expected_allele_length")
+                for sequence, (forward, reverse) in sequences.items():
                     if not seq_pass_filt(sequence, forward + reverse, minimum, expected_length):
                         if marker not in aggregates:
                             aggregates[marker] = [0, 0]
@@ -246,10 +290,10 @@ class TSSV:
                         aggregates[marker][1] += reverse
 
         # Filter out sequences with low read counts and invalid bases.
-        for marker, sequences in self.sequences.items():
-            expected_length = self.library.get("expected_length", {}).get(marker)
+        for marker in tuple(self.sequences):
+            expected_length = self.library.get_range(marker).get_option("expected_allele_length")
             self.sequences[marker] = {sequence: counts
-                for sequence, counts in sequences.items()
+                for sequence, counts in self.sequences[marker].items()
                 if seq_pass_filt(sequence, sum(counts), minimum, expected_length)}
 
         # Add aggregate rows if the user requested so.
@@ -280,7 +324,8 @@ class TSSV:
             for total, seq in sorted(((sum(sequences[s]), s) for s in sequences), reverse=True):
                 line = "\t".join(map(str, [
                     marker,
-                    ensure_sequence_format(seq, seqformat, "raw", self.library, marker),
+                    ensure_sequence_format(seq, seqformat, from_format="raw", library=self.library,
+                        marker=marker),
                     total] + sequences[seq])) + "\n"
                 outfile.write(line)
                 if self.outfiles:
@@ -294,7 +339,7 @@ class TSSV:
                             "tLeft", "fLeft", "rLeft", "tRight", "fRight", "rRight")) + "\n"
         if self.outfiles:
             self.outfiles["statistics"].write(header)
-        outfile.write(header)
+        try_write_pipe(outfile, header)
         for marker in sorted(self.counters):
             line = "\t".join(map(str, (
                 marker,
@@ -310,85 +355,160 @@ class TSSV:
                 self.counters[marker]["rRight"]))) + "\n"
             if self.outfiles:
                 self.outfiles["statistics"].write(line)
-            outfile.write(line)
+            try_write_pipe(outfile, line)
 
         line = "\ntotal reads\t%i\nunrecognised reads\t%i\n" % (self.total_reads,self.unrecognised)
         if self.outfiles:
             self.outfiles["statistics"].write(line)
-        outfile.write(line)
+        try_write_pipe(outfile, line)
     #write_statistics_table
+
+
+    def close_output_files(self):
+        if self.outfiles:
+            for key, value in self.outfiles.items():
+                if key == "markers":
+                    for markerfiles in value.values():
+                        for markerfile in markerfiles.values():
+                            markerfile.close()
+                else:
+                    value.close()
+    #close_output_files
 #TSSV
 
 
 def prepare_output_dir(dir, markers, file_format):
     # Create output directories.
-    os.makedirs(dir)
     for marker in markers:
-        os.mkdir(os.path.join(dir, marker))
+        os.makedirs(os.path.join(dir, marker), exist_ok=True)
 
     # Open output files.
+    open_seq_out = gzip.open if file_format.endswith(".gz") else open
     return {
-        "sequences": open(os.path.join(dir, "sequences.csv"), "w"),
-        "statistics": open(os.path.join(dir, "statistics.csv"), "w"),
-        "unknown": open(os.path.join(dir, "unknown.f" + file_format[-1]), "w"),
+        "sequences": open(os.path.join(dir, "sequences.csv"), "wt", encoding="UTF-8"),
+        "statistics": open(os.path.join(dir, "statistics.csv"), "wt", encoding="UTF-8"),
+        "unknown": open_seq_out(os.path.join(dir, "unknown" + file_format), "wt", encoding="UTF-8"),
         "markers": {
             marker: {
-                "sequences": open(os.path.join(dir, marker, "sequences.csv"), "w"),
-                "paired": open(os.path.join(dir, marker, "paired.f" + file_format[-1]), "w"),
-                "noend": open(os.path.join(dir, marker, "noend.f" + file_format[-1]), "w"),
-                "nostart": open(os.path.join(dir, marker, "nostart.f" + file_format[-1]), "w"),
+                "sequences": open(os.path.join(dir, marker, "sequences.csv"), "wt",
+                    encoding="UTF-8"),
+                "paired": open_seq_out(os.path.join(dir, marker, "paired" + file_format), "wt",
+                    encoding="UTF-8"),
+                "noend": open_seq_out(os.path.join(dir, marker, "noend" + file_format), "wt",
+                    encoding="UTF-8"),
+                "nostart": open_seq_out(os.path.join(dir, marker, "nostart" + file_format), "wt",
+                    encoding="UTF-8"),
             } for marker in markers
         }
     }
 #prepare_output_dir
 
 
-def align_pair(reference, reference_rc, pair, indel_score=1):
-    left_dist, left_pos = align(reference, pair[0], indel_score)
-    right_dist, right_pos = align(reference_rc, pair[1], indel_score)
-    return (left_dist, left_pos), (right_dist, len(reference) - right_pos)
+def align_pair(reference, reference_rc, pair, indel_score):
+    left_dist, left_pos = align(
+        reference[pair[0][1]], pair[0][0], indel_score, bitwise=pair[0][1])
+    right_dist, right_pos = align(
+        reference_rc[pair[1][1]], pair[1][0], indel_score, bitwise=pair[1][1])
+    return (left_dist, left_pos), (right_dist, len(reference[0]) - right_pos)
 #align_pair
 
 
-def process_sequence(tssv_library, indel_score, seq):
+def relative_distance(reference, sequence):
+    """Return (pct_changed, -len(reference))."""
+    len_reference = len(reference)
+    if len_reference > len(sequence):
+        return (align(reference, sequence, 1, global_align=1)[0] / len_reference, -len_reference)
+    return (align(sequence, reference, 1, global_align=1)[0] / len_reference, -len_reference)
+#relative_distance
+
+
+def prune_matched_ranges(tssv_library, matched_ranges):
+    """Removes overlapping ranges from the provided list."""
+    if len(matched_ranges) < 2:
+        return
+
+    matched_ranges.sort()
+    group_scores = []
+    group_start = 0
+    i = 1
+    while i <= len(matched_ranges):
+        if i == len(matched_ranges) or all(
+                (matched_ranges[i][1]-matched_ranges[i][0] < y[1]-y[0] and
+                    matched_ranges[i][0] + (matched_ranges[i][1]-matched_ranges[i][0])//2 >= y[1])
+                or (matched_ranges[i][1]-matched_ranges[i][0] >= y[1]-y[0] and
+                    y[1] - (y[1]-y[0])//2 <= matched_ranges[i][0])
+                    for y in matched_ranges[group_start:i]):
+            # End of group with at least 50% pairwise overlap.
+            if i - group_start > 1:
+                # Overlapping group; eliminate the worst fit iteratively.
+                if not group_scores:
+                    group_scores = [relative_distance(tssv_library[marker][2], seq)
+                        for start, end, seq, strand, marker in matched_ranges]
+                worst_range_index = group_scores.index(max(group_scores[:i - group_start]))
+                del matched_ranges[group_start + worst_range_index]
+                del group_scores[worst_range_index]
+                i = group_start + 1
+                continue
+            # Group consisted of a single range; new group starts here.
+            group_scores = group_scores[1:]
+            group_start = i
+        i += 1
+#prune_matched_ranges
+
+
+def process_sequence(tssv_library, indel_score, has_iupac, seq):
     """Find markers in sequence."""
     seqs = (seq, reverse_complement(seq))
-    seqs_up = map(str.upper, seqs)
-    results = []
-    for marker, (pair, thresholds) in tssv_library.items():
+    seqs_up = tuple((x, x.translate(IUPAC_BITS) if has_iupac else None)
+        for x in (seqs[0].upper(), seqs[1].upper()))
+    matched_ranges = []
+    marker_matches = {}
+    for marker, (pair, thresholds, refseq) in tssv_library.items():
         algn = (
             align_pair(seqs_up[0], seqs_up[1], pair, indel_score),
             align_pair(seqs_up[1], seqs_up[0], pair, indel_score))
         matches = 0
         if algn[0][0][0] <= thresholds[0]:
             # Left marker was found in forward sequence
-            cutout = seqs[0][max(0, algn[0][0][1]-len(pair[0])):algn[0][0][1]]
+            cutout = seqs[0][max(0, algn[0][0][1] - len(pair[0][0])) : algn[0][0][1]]
             if cutout.lower() != cutout:
-                matches += 1
+                matches += 0b0001
         if algn[0][1][0] <= thresholds[1]:
             # Right marker was found in forward sequence.
-            cutout = seqs[0][algn[0][1][1] : algn[0][1][1]+len(pair[1])]
+            cutout = seqs[0][algn[0][1][1] : algn[0][1][1] + len(pair[1][0])]
             if cutout.lower() != cutout:
-                matches += 2
+                matches += 0b0010
         if algn[1][0][0] <= thresholds[0]:
             # Left marker was found in reverse sequence
-            cutout = seqs[1][max(0, algn[1][0][1]-len(pair[0])):algn[1][0][1]]
+            cutout = seqs[1][max(0, algn[1][0][1] - len(pair[0][0])) : algn[1][0][1]]
             if cutout.lower() != cutout:
-                matches += 4
+                matches += 0b0100
         if algn[1][1][0] <= thresholds[1]:
             # Right marker was found in reverse sequence.
-            cutout = seqs[1][algn[1][1][1] : algn[1][1][1]+len(pair[1])]
+            cutout = seqs[1][algn[1][1][1] : algn[1][1][1] + len(pair[1][0])]
             if cutout.lower() != cutout:
-                matches += 8
-        results.append((marker, matches,
+                matches += 0b1000
+        if (matches & 0b0011) == 0b0011 and algn[0][0][1] < algn[0][1][1]:
             # Matched pair in forward sequence.
-            seqs[0][algn[0][0][1] : algn[0][1][1]] if
-                (matches & 3) == 3 and algn[0][0][1] < algn[0][1][1]
-                else None,
+            matched_ranges.append((algn[0][0][1], algn[0][1][1],
+                seqs[0][algn[0][0][1] : algn[0][1][1]], 0, marker))
+        if (matches & 0b1100) == 0b1100 and algn[1][0][1] < algn[1][1][1]:
             # Matched pair in reverse sequence.
-            seqs[1][algn[1][0][1] : algn[1][1][1]] if
-                (matches & 12) == 12 and algn[1][0][1] < algn[1][1][1]
-                else None))
+            matched_ranges.append((len(seq) - algn[1][1][1], len(seq) - algn[1][0][1],
+                seqs[1][algn[1][0][1] : algn[1][1][1]], 1, marker))
+        marker_matches[marker] = matches
+
+    # Eliminate overlapping matched pairs.
+    prune_matched_ranges(tssv_library, matched_ranges)
+
+    # Compile the results.
+    results = []
+    for marker, matches in marker_matches.items():
+        forward = [match[2] for match in matched_ranges if match[3] == 0 and match[4] == marker]
+        reverse = [match[2] for match in matched_ranges if match[3] == 1 and match[4] == marker]
+        results.append((marker, matches,
+            forward[0] if forward else None,
+            reverse[0] if reverse else None))
     return results
 #process_sequence
 
@@ -400,16 +520,17 @@ def seq_pass_filt(sequence, reads, threshold, explen=None):
 #seq_pass_filt
 
 
-def worker(tssv_library, indel_score, task_queue, done_queue):
+def worker(tssv_library, indel_score, has_iupac, task_queue, done_queue):
     """
     Read sequences from task_queue, write findings to done_queue.
     """
     for task in iter(task_queue.get, None):
-        done_queue.put(tuple((seq, process_sequence(tssv_library, indel_score, seq)) for seq in task))
+        done_queue.put(tuple(
+            (seq, process_sequence(tssv_library, indel_score, has_iupac, seq)) for seq in task))
 #worker
 
 
-def feeder(input, tssv_library, indel_score, workers, chunksize, done_queue):
+def feeder(input, tssv_library, indel_score, has_iupac, workers, chunksize, done_queue):
     """
     Start worker processes, feed them sequences from input and have them
     write their results to done_queue.
@@ -418,13 +539,13 @@ def feeder(input, tssv_library, indel_score, workers, chunksize, done_queue):
     processes = []
     for i in range(workers):
         process = Process(target=worker,
-            args=(tssv_library, indel_score, task_queue, done_queue))
+            args=(tssv_library, indel_score, has_iupac, task_queue, done_queue))
         process.daemon = True
         process.start()
         processes.append(process)
-    while 1:
+    while True:
         # Sending chunks of reads to the workers.
-        task = tuple(it.islice(input, chunksize))
+        task = tuple(itertools.islice(input, chunksize))
         if not task:
             break
         task_queue.put(task)
@@ -445,17 +566,29 @@ def write_sequence_record(outfile, record):
 #write_sequence_record
 
 
-def init_sequence_file_read(infile):
+def init_sequence_file_read(rawfile):
     """
     Return a 2-tuple with "fasta" or "fastq" and a generator that
     generates tuples of (header, sequence) from a FastA stream or
     tuples of (header, sequence, quality) from a FastQ stream.
     """
+    magic = rawfile.peek(2)[:2]
+    if magic == b"\x1f\x8b":
+        # GZipped input file.
+        infile = gzip.open(rawfile, "rt", encoding="UTF-8")
+        filetype = ".gz"
+    elif magic and magic[0] in b">@":
+        # FastA/FastQ input file.
+        infile = io.TextIOWrapper(rawfile, encoding="UTF-8")
+        filetype = ".fa" if magic[0] == b">" else ".fq"
+    else:
+        raise ValueError("Input file is not a GZip, FastQ or FastA file")
+
     firstchar = infile.read(1)
-    if not firstchar:
-        return
-    if firstchar not in ">@":
-        raise ValueError("Input file is not a FastQ or FastA file")
+    if not firstchar or firstchar not in ">@":
+        raise ValueError("Input GZip file does not contain a FastQ or FastA file")
+    if filetype == ".gz":
+        filetype = ".fa.gz" if firstchar == ">" else ".fq.gz"
 
     def genreads():
         state = 0
@@ -484,29 +617,40 @@ def init_sequence_file_read(infile):
                 state = 1
         yield (header, seq) if state == 1 else (header, seq, qual)
 
-    return "fasta" if firstchar == ">" else "fastq", genreads()
+    return filetype, genreads()
 #init_sequence_file_read
 
 
-def run_tssv_lite(infile, outfile, reportfile, library, seqformat,
-                  threshold, minimum, aggregate_filtered,
-                  missing_marker_action, dirname, indel_score, workers, no_deduplicate):
-    tssv = TSSV(library, threshold, indel_score, dirname, workers, not no_deduplicate, infile)
+def run_tssv_lite(infile, outfile, reportfile, library, flank_length, seqformat, threshold,
+                  minimum, aggregate_filtered, missing_marker_action, dirname, indel_score,
+                  workers, no_deduplicate):
+    tssv = TSSV(library, flank_length, threshold, indel_score, dirname, workers,
+        not no_deduplicate, infile)
     tssv.process_file()
     tssv.filter_sequences(aggregate_filtered, minimum, missing_marker_action)
 
     try:
         tssv.write_sequence_tables(outfile, seqformat)
         tssv.write_statistics_table(reportfile)
+        tssv.close_output_files()
     except IOError as e:
-        if e.errno != EPIPE:
-            raise
+        if e.errno == EPIPE:
+            try:
+                try_write_pipe(reportfile, "Stopped early because the output was closed.\n")
+            except:
+                pass
+            return
+        raise
 #run_tssv_lite
 
 
 def add_arguments(parser):
-    add_sequence_format_args(parser, "raw", False, True)
-    add_input_output_args(parser, True, False, True)
+    add_sequence_format_args(parser, default_format="raw", force=False, require_library=True)
+    add_input_output_args(parser, single_in=True, batch_support=False, report_out=True)
+    parser.add_argument("-L", "--flank-length", metavar="N", type=pos_int_arg,
+        default=_DEF_FLANK_LENGTH,
+        help="length of anchor (flanking) sequences to use, if not specified in the library file "
+             "(default: %(default)s)")
     parser.add_argument("-D", "--dir",
         help="output directory for verbose output; when given, a subdirectory "
              "will be created for each marker, each with a separate "
@@ -514,49 +658,44 @@ def add_arguments(parser):
              "unrecognised reads (unknown.fa), recognised reads "
              "(Marker/paired.fa), and reads that lack one of the flanks of a "
              "marker (Marker/noend.fa and Marker/nostart.fa)")
-    parser.add_argument("-T", "--num-threads", metavar="THREADS",
-        type=pos_int_arg, default=1,
+    parser.add_argument("-T", "--num-threads", metavar="THREADS", type=pos_int_arg, default=1,
         help="number of worker threads to use (default: %(default)s)")
     parser.add_argument("-X", "--no-deduplicate", action="store_true",
         help="disable deduplication of reads; by setting this option, memory usage will be "
              "reduced in expense of longer running time")
     filtergroup = parser.add_argument_group("filtering options")
-    filtergroup.add_argument("-m", "--mismatches", type=float,
-        default=_DEF_MISMATCHES,
-        help="number of mismatches per nucleotide to allow in flanking "
-             "sequences (default: %(default)s)")
+    filtergroup.add_argument("-m", "--mismatches", type=float, default=_DEF_MISMATCHES,
+        help="number of mismatches (per nucleotide of flanking sequence if less than 1, else "
+             "absolute) to allow in flanking sequences, rounded upward (default: %(default)s)")
     filtergroup.add_argument("-n", "--indel-score", metavar="N",
         type=pos_int_arg, default=_DEF_INDEL_SCORE,
-        help="insertions and deletions in the flanking sequences are "
-             "penalised this number of times more heavily than mismatches "
-             "(default: %(default)s)")
+        help="insertions and deletions in the flanking sequences are penalised this number of "
+             "times more heavily than mismatches (default: %(default)s)")
     filtergroup.add_argument("-a", "--minimum", metavar="N", type=pos_int_arg,
         default=_DEF_MINIMUM,
-        help="report only sequences with this minimum number of reads "
-             "(default: %(default)s)")
-    filtergroup.add_argument("-A", "--aggregate-filtered", action="store_true",
-        help="if specified, sequences that have been filtered (as per the "
-             "-a/--minimum option, the expected_allele_length section in the "
-             "library file, as well as all sequences with ambiguous bases) "
-             "will be aggregated per marker and reported as 'Other sequences'")
+        help="report only sequences with this minimum number of reads (default: %(default)s)")
+    filtergroup.add_argument("-B", "--no-aggregate-filtered", dest="aggregate_filtered",
+        action="store_false",
+        help="by default, sequences that have been filtered (as per the -a/--minimum option, the "
+             "expected_allele_length section in the library file, as well as all sequences with "
+             "ambiguous bases) will be aggregated per marker and reported as 'Other sequences'; "
+             "specify this option to remove such sequences entirely")
     filtergroup.add_argument("-M", "--missing-marker-action", metavar="ACTION",
-        choices=("include", "exclude", "halt"),
-        default="include",
+        choices=("include", "exclude", "halt"), default="include",
         help="action to take when no sequences are linked to a marker: one of "
              "%(choices)s (default: %(default)s)")
 #add_arguments
 
 
 def run(args):
-    files = get_input_output_files(args, True, False)
+    files = get_input_output_files(args, single_in=True, batch_support=False)
     if not files:
-        raise ValueError("please specify an input file, or pipe in the output "
-                         "of another program")
-    infile = sys.stdin if files[0] == "-" else open(files[0], "r")
-    run_tssv_lite(infile, files[1], args.report, args.library,
+        raise ValueError("please specify an input file, or pipe in the output of another program")
+    infile = sys.stdin.buffer if files[0] == "-" else open(files[0], "rb")
+    run_tssv_lite(infile, files[1], args.report, args.library, args.flank_length,
                   args.sequence_format, args.mismatches, args.minimum,
                   args.aggregate_filtered, args.missing_marker_action,
                   args.dir, args.indel_score, args.num_threads, args.no_deduplicate)
-    if infile != sys.stdin:
+    if infile != sys.stdin.buffer:
         infile.close()
 #run
