@@ -1,10 +1,10 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 
 #
-# Copyright (C) 2017 Jerry Hoogenboom
+# Copyright (C) 2021 Jerry Hoogenboom
 #
-# This file is part of FDSTools, data analysis tools for Next
-# Generation Sequencing of forensic DNA markers.
+# This file is part of FDSTools, data analysis tools for Massively
+# Parallel Sequencing of forensic DNA markers.
 #
 # FDSTools is free software: you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -45,17 +45,17 @@ specifically to visualise the contents of this file.  To this end, the
 entire file contents are embedded in the generated visualisation.
 """
 import argparse
-import sys
 import json
+import os.path
+import pkgutil
 import re
-import os
+import sys
 
-from pkg_resources import resource_stream, resource_string
 from errno import EPIPE
 
-from ..lib import pos_int_arg
+from ..lib.cli import pos_int_arg
 
-__version__ = "1.0.4"
+__version__ = "1.1.0"
 
 
 # Default values for parameters are specified below.
@@ -102,12 +102,12 @@ _DEF_THRESHOLD_CORRECTION_ALLELE = 0
 _DEF_THRESHOLD_RECOVERY_ALLELE = 0
 
 # Default minimum number of reads per strand to mark as allele in Samplevis.
-# This value can be overridden by the -b command line option.
-_DEF_THRESHOLD_ORIENTATION_ALLELE = 1
+# This value can be overridden by the -Z command line option.
+_DEF_THRESHOLD_ORIENTATION_ALLELE = 0
 
 # Default percentage of reads on one strand to mark as bias.
 # This value can be overridden by the -B command line option.
-_DEF_THRESHOLD_BIAS = 25.0
+_DEF_THRESHOLD_BIAS = 0.0
 
 # Default width of bars in bar graphs, in pixels.
 # This value can be overridden by the -b command line option.
@@ -147,19 +147,17 @@ _DEF_UNIT = ""
 _DEF_DATA_FILENAME = "data.csv"
 
 
-
 _PAT_LIBRARIES = re.compile("<!--\s*BEGIN_LIBRARIES\s*-->\s*(.*?)\s*"
                             "<!--\s*END_LIBRARIES\s*-->", flags=re.DOTALL)
 _PAT_LOAD_SCRIPT = re.compile("<!--\s*BEGIN_LOAD_SCRIPT\s*-->\s*(.*?)\s*"
                               "<!--\s*END_LOAD_SCRIPT\s*-->", flags=re.DOTALL)
 
 _SCRIPT_BEGIN = '<script type="text/javascript">'
-_SCRIPT_END = '</script>'
+_SCRIPT_END = "</script>"
 
 _EXTERNAL_LIBRARIES = ("vis/d3.min.js", "vis/vega.min.js")
 
 
-
 def set_signal_value(spec, signalname, value):
     if "signals" not in spec:
         return False
@@ -185,18 +183,15 @@ def set_axis_scale(spec, scalename, value):
 #set_axis_scale
 
 
-def create_visualisation(vistype, infile, infile2, outfile, vega, online, tidy,
-                         min_abs, min_pct_of_max, min_pct_of_sum,
-                         min_per_strand, bias_threshold, bar_width, padding,
-                         marker, width, height, log_scale, repeat_unit,
-                         no_alldata, no_aggregate, no_ce_length_sort,
-                         max_seq_len, jitter, title, allele_min_abs,
-                         allele_min_pct_of_max, allele_min_pct_of_sum,
-                         allele_min_correction, allele_min_recovery,
-                         allele_min_per_strand):
+def create_visualisation(vistype, infile, infile2, outfile, vega, tidy, min_abs, min_pct_of_max,
+                         min_pct_of_sum, min_per_strand, bias_threshold, bar_width, padding,
+                         marker, width, height, log_scale, repeat_unit, no_alldata, no_aggregate,
+                         no_ce_length_sort, max_seq_len, jitter, title, allele_min_abs,
+                         allele_min_pct_of_max, allele_min_pct_of_sum, allele_min_correction,
+                         allele_min_recovery, allele_min_per_strand):
     # Get graph spec.
-    spec = json.load(resource_stream(
-        "fdstools", "vis/%svis/%svis.json" % (vistype, vistype)))
+    spec = json.loads(
+        pkgutil.get_data("fdstools", "vis/%svis/%svis.json" % (vistype, vistype)).decode())
     if infile is not None:
         # Embed the given data file.
         spec["data"][0]["values"] = infile.read()
@@ -239,16 +234,11 @@ def create_visualisation(vistype, infile, infile2, outfile, vega, online, tidy,
         set_signal_value(spec, "sort_str_by_length", not no_ce_length_sort)
         set_signal_value(spec, "max_seq_len", max_seq_len)
         set_signal_value(spec, "allele_amplitude_threshold", allele_min_abs)
-        set_signal_value(spec, "allele_amplitude_pct_threshold",
-            allele_min_pct_of_max)
-        set_signal_value(spec, "allele_amplitude_markerpct_threshold",
-            allele_min_pct_of_sum)
-        set_signal_value(spec, "allele_correction_threshold",
-            allele_min_correction)
-        set_signal_value(spec, "allele_recovery_threshold",
-            allele_min_recovery)
-        set_signal_value(spec, "allele_orientation_threshold",
-            allele_min_per_strand)
+        set_signal_value(spec, "allele_amplitude_pct_threshold", allele_min_pct_of_max)
+        set_signal_value(spec, "allele_amplitude_markerpct_threshold", allele_min_pct_of_sum)
+        set_signal_value(spec, "allele_correction_threshold", allele_min_correction)
+        set_signal_value(spec, "allele_recovery_threshold", allele_min_recovery)
+        set_signal_value(spec, "allele_orientation_threshold", allele_min_per_strand)
 
     # Apply axis scale settings.
     if vistype != "stuttermodel" and vistype != "allele":
@@ -280,7 +270,7 @@ def create_visualisation(vistype, infile, infile2, outfile, vega, online, tidy,
         return
 
     # Creating a fully self-contained HTML visualisation instead.
-    html = resource_string("fdstools", "vis/%svis/index.html" % vistype)
+    html = pkgutil.get_data("fdstools", "vis/%svis/index.html" % vistype).decode()
     match = _PAT_LOAD_SCRIPT.search(html)
     if match:
         html = "".join([
@@ -296,26 +286,22 @@ def create_visualisation(vistype, infile, infile2, outfile, vega, online, tidy,
             _SCRIPT_END,
             html[match.end(1):]])
 
-    if not online:
-        # Replace external libraries with inline libraries.
-        match = _PAT_LIBRARIES.search(html)
-        if match:
-            parts = [html[:match.start(1)]]
-            for library in _EXTERNAL_LIBRARIES:
-                parts += [_SCRIPT_BEGIN,
-                          resource_string("fdstools", library),
-                          _SCRIPT_END]
-            parts.append(html[match.end(1):])
-            html = "".join(parts)
+    # Replace external libraries with inline libraries.
+    match = _PAT_LIBRARIES.search(html)
+    if match:
+        parts = [html[:match.start(1)]]
+        for library in _EXTERNAL_LIBRARIES:
+            parts += [_SCRIPT_BEGIN, pkgutil.get_data("fdstools", library).decode(), _SCRIPT_END]
+        parts.append(html[match.end(1):])
+        html = "".join(parts)
 
     outfile.write(html)
 #create_visualisation
 
 
 def add_arguments(parser):
-    parser.add_argument('type', metavar="TYPE",
-        choices=("sample", "profile", "bgraw", "stuttermodel", "allele",
-            "bganalyse"),
+    parser.add_argument("type", metavar="TYPE",
+        choices=("sample", "profile", "bgraw", "stuttermodel", "allele", "bganalyse"),
         help="the type of data to visualise; use 'sample' to visualise "
              "sample data files and bgcorrect output; use 'profile' to "
              "visualise background noise profiles obtained with bgestimate, "
@@ -325,77 +311,66 @@ def add_arguments(parser):
              "stuttermodel; 'bganalyse' to visualise data obtained from "
              "bganalyse; use 'allele' to visualise the allele list obtained "
              "from allelefinder")
-    parser.add_argument('infile', metavar="IN", nargs="?",
+    parser.add_argument("infile", metavar="IN", nargs="?", default=sys.stdin,
         help="file containing the data to embed in the visualisation file; if "
              "not specified, HTML visualisation files will contain a file "
              "selection control, and Vega visualisation files will load data "
              "from a file called '%s'" % _DEF_DATA_FILENAME)
-    parser.add_argument('outfile', metavar="OUT", nargs="?",
-        type=argparse.FileType('w'),
-        default=sys.stdout,
+    parser.add_argument("outfile", metavar="OUT", nargs="?",
+        type=argparse.FileType("tw", encoding="UTF-8"), default=sys.stdout,
         help="file to write output to (default: write to stdout)")
-    parser.add_argument('-V', '--vega', action="store_true",
+    parser.add_argument("-V", "--vega", action="store_true",
         help="by default, a full-featured HTML file offering an interactive "
              "visualisation is created; if this option is specified, only a "
              "bare Vega graph specification (JSON file) is produced instead")
-    parser.add_argument('-O', '--online', action="store_true",
-        help="when generating an HTML visualisation file, required JavaScript "
-             "libraries (D3 and Vega) are embedded in the file; if this "
-             "option is specified, the HTML file will instead link to these "
-             "libraries on the Internet, thereby always using the latest "
-             "versions of D3 and Vega")
-    parser.add_argument('-t', '--tidy', action="store_true",
+    parser.add_argument("-t", "--tidy", action="store_true",
         help="tidily indent the generated JSON")
-    parser.add_argument('-T', '--title',
+    parser.add_argument("-T", "--title",
         help="prepend the given value to the title of HTML visualisations "
              "(default: prepend name of data file if given)")
 
     visgroup = parser.add_argument_group("visualisation options",
-        description="words in [brackets] indicate applicable visualisation "
-                    "types")
-    visgroup.add_argument('-n', '--min-abs', metavar="N", type=float,
-        default=_DEF_THRESHOLD_ABS,
+        description="words in [brackets] indicate applicable visualisation types")
+    visgroup.add_argument("-n", "--min-abs", metavar="N", type=float, default=_DEF_THRESHOLD_ABS,
         help="[sample, profile, bgraw] only show sequences with this minimum "
              "number of reads (default: %(default)s)")
-    visgroup.add_argument('-m', '--min-pct-of-max', metavar="PCT", type=float,
+    visgroup.add_argument("-m", "--min-pct-of-max", metavar="PCT", type=float,
         default=_DEF_THRESHOLD_PCT_OF_MAX,
         help="[sample, profile, bgraw] for sample: only show sequences with "
              "at least this percentage of the number of reads of the highest "
              "allele of a marker; for profile and bgraw: at least this "
              "percentage of the true allele (default: %(default)s)")
-    visgroup.add_argument('-S', '--min-pct-of-sum', metavar="PCT", type=float,
+    visgroup.add_argument("-S", "--min-pct-of-sum", metavar="PCT", type=float,
         default=_DEF_THRESHOLD_PCT_OF_SUM,
         help="[sample] only show sequences with at least this percentage of "
              "the total number of reads of a marker (default: %(default)s)")
-    visgroup.add_argument('-s', '--min-per-strand', metavar="N", type=float,
+    visgroup.add_argument("-s", "--min-per-strand", metavar="N", type=float,
         default=_DEF_THRESHOLD_ORIENTATION,
         help="[sample] only show sequences with this minimum number of reads "
              "for both orientations (forward/reverse) (default: %(default)s)")
-    visgroup.add_argument('-B', '--bias-threshold', metavar="N", type=float,
+    visgroup.add_argument("-B", "--bias-threshold", metavar="N", type=float,
         default=_DEF_THRESHOLD_BIAS,
         help="[sample] mark sequences that have less than this percentage of "
              "reads on one strand (default: %(default)s)")
-    visgroup.add_argument('-c', '--no-ce-length-sort', action="store_true",
+    visgroup.add_argument("-c", "--no-ce-length-sort", action="store_true",
         help="[sample] if specified, do not sort STR alleles by length")
-    visgroup.add_argument('-M', '--marker', metavar="MARKER",
-        default=_DEF_MARKER_NAME,
+    visgroup.add_argument("-M", "--marker", metavar="MARKER", default=_DEF_MARKER_NAME,
         help="[sample, profile, bgraw, stuttermodel, bganalyse] only show "
              "graphs for the markers that contain the given value in their "
              "name; separate multiple values with spaces; prepend any value "
              "with '=' for an exact match (default: show all markers)")
-    visgroup.add_argument('-U', '--repeat-unit', metavar="UNIT",
-        default=_DEF_UNIT,
+    visgroup.add_argument("-U", "--repeat-unit", metavar="UNIT", default=_DEF_UNIT,
         help="[stuttermodel] only show graphs for the repeat units that "
              "contain the given value; separate multiple values with spaces; "
              "prepend any value with '=' for an exact match (default: show "
              "all repeat units)")
-    visgroup.add_argument('-A', '--no-alldata', action="store_true",
+    visgroup.add_argument("-A", "--no-alldata", action="store_true",
         help="[stuttermodel] if specified, show only marker-specific fits")
-    visgroup.add_argument('-a', '--no-aggregate', action="store_true",
+    visgroup.add_argument("-a", "--no-aggregate", action="store_true",
         help="[sample] if specified, do not replace filtered sequences with a "
              "per-marker aggregate 'Other sequences' entry")
-    visgroup.add_argument('-I', '--input2', dest="infile2", metavar="FILE",
-        type=argparse.FileType("r"),
+    visgroup.add_argument("-I", "--input2", dest="infile2", metavar="FILE",
+        type=argparse.FileType("tr", encoding="UTF-8"),
         help="[profile, stuttermodel] raw data points file to overlay on the "
              "background noise profiles or stutter model graphs (as obtained "
              "from bghomraw or the -r/--raw-outfile option of stuttermodel); "
@@ -403,99 +378,87 @@ def add_arguments(parser):
              "selection control")
 
     dispgroup = parser.add_argument_group("display options")
-    dispgroup.add_argument('-L', '--log-scale', action="store_true",
+    dispgroup.add_argument("-L", "--log-scale", action="store_true",
         help="[sample, profile, bgraw, bganalyse] use logarithmic scale (for "
-             "sample and bganalyse: square root scale) instead of linear "
-             "scale")
-    dispgroup.add_argument('-b', '--bar-width', metavar="N", type=pos_int_arg,
+             "sample and bganalyse: square root scale) instead of linear scale")
+    dispgroup.add_argument("-b", "--bar-width", metavar="N", type=pos_int_arg,
         default=_DEF_BAR_WIDTH,
         help="[sample, profile, bgraw, bganalyse] width of the bars in pixels "
              "(default: %(default)s)")
-    dispgroup.add_argument('-p', '--padding', metavar="N", type=pos_int_arg,
+    dispgroup.add_argument("-p", "--padding", metavar="N", type=pos_int_arg,
         default=_DEF_SUBGRAPH_PADDING,
         help="[sample, profile, bgraw, stuttermodel] amount of padding (in "
-             "pixels) between graphs of different markers/alleles (default: "
-             "%(default)s)")
-    dispgroup.add_argument('-w', '--width', metavar="N", type=pos_int_arg,
-        default=_DEF_WIDTH,
+             "pixels) between graphs of different markers/alleles (default: %(default)s)")
+    dispgroup.add_argument("-w", "--width", metavar="N", type=pos_int_arg, default=_DEF_WIDTH,
         help="[sample, profile, bgraw, stuttermodel, bganalyse, allele] width "
              "of the graph area in pixels (default: %(default)s)")
-    dispgroup.add_argument('-H', '--height', metavar="N", type=pos_int_arg,
-        default=_DEF_HEIGHT,
-        help="[stuttermodel, allele] height of the graph area in pixels "
-             "(default: %(default)s)")
-    dispgroup.add_argument('-x', '--max-seq-len', metavar="N", type=pos_int_arg,
+    dispgroup.add_argument("-H", "--height", metavar="N", type=pos_int_arg, default=_DEF_HEIGHT,
+        help="[stuttermodel, allele] height of the graph area in pixels (default: %(default)s)")
+    dispgroup.add_argument("-x", "--max-seq-len", metavar="N", type=pos_int_arg,
         default=_DEF_MAX_SEQ_LEN,
         help="[sample] truncate long sequences to this number of characters "
              "(default: %(default)s)")
-    dispgroup.add_argument('-j', '--jitter', metavar="N", type=float,
-        default=_DEF_JITTER,
+    dispgroup.add_argument("-j", "--jitter", metavar="N", type=float, default=_DEF_JITTER,
         help="[stuttermodel] apply this amount of jitter to raw data points "
              "(between 0 and 1, default: %(default)s)")
 
     allelegroup = parser.add_argument_group("allele calling options",
         "for sample visualisations only; sequences that match the -C or -Y "
-        "option (or both) and all of the other settings are marked as "
-        "'allele'")
-    allelegroup.add_argument('-N', '--allele-min-abs', metavar="N", type=float,
+        "option (or both) and all of the other settings are marked as 'allele'")
+    allelegroup.add_argument("-N", "--allele-min-abs", metavar="N", type=float,
         default=_DEF_THRESHOLD_ABS_ALLELE,
         help="the minimum number of reads (default: %(default)s)")
-    allelegroup.add_argument('-X', '--allele-min-pct-of-max', metavar="PCT",
+    allelegroup.add_argument("-X", "--allele-min-pct-of-max", metavar="PCT",
         type=float, default=_DEF_THRESHOLD_PCT_OF_MAX_ALLELE,
         help="the minimum percentage of reads w.r.t. the highest allele of "
              "the marker (default: %(default)s)")
-    allelegroup.add_argument('-Q', '--allele-min-pct-of-sum', metavar="PCT",
+    allelegroup.add_argument("-Q", "--allele-min-pct-of-sum", metavar="PCT",
         type=float, default=_DEF_THRESHOLD_PCT_OF_SUM_ALLELE,
         help="the minimum percentage of reads w.r.t. the marker's total "
              "number of reads (default: %(default)s)")
-    allelegroup.add_argument('-C', '--allele-min-correction', metavar="N",
+    allelegroup.add_argument("-C", "--allele-min-correction", metavar="N",
         type=float, default=_DEF_THRESHOLD_CORRECTION_ALLELE,
         help="the minimum change in read count due to correction by e.g., "
              "bgcorrect (default: %(default)s)")
-    allelegroup.add_argument('-Y', '--allele-min-recovery', metavar="N",
+    allelegroup.add_argument("-Y", "--allele-min-recovery", metavar="N",
         type=float, default=_DEF_THRESHOLD_RECOVERY_ALLELE,
         help="the minimum number of reads that was recovered thanks to "
              "noise correction (by e.g., bgcorrect), as a percentage of the "
              "total number of reads after correction (default: %(default)s)")
-    allelegroup.add_argument('-Z', '--allele-min-per-strand', metavar="N",
+    allelegroup.add_argument("-Z", "--allele-min-per-strand", metavar="N",
         type=float, default=_DEF_THRESHOLD_ORIENTATION_ALLELE,
-        help="the minimum number of reads in both orientations (default: "
-             "%(default)s)")
+        help="the minimum number of reads in both orientations (default: %(default)s)")
 #add_arguments
 
 
 def run(args):
     if args.infile == "-":
-        args.infile = None if sys.stdin.isatty() else sys.stdin
-    if (args.infile is not None and args.outfile == sys.stdout
-            and not os.path.exists(args.infile)):
+        args.infile = sys.stdin
+    if args.infile != sys.stdin and args.outfile == sys.stdout and not os.path.exists(args.infile):
         # One filename given, and it does not exist.  Assume outfile.
-        args.outfile = open(args.infile, 'w')
+        args.outfile = open(args.infile, "wt", encoding="UTF-8")
+        args.infile = sys.stdin
+    elif args.outfile.isatty():
+        raise ValueError("Please specify a file name to write the %s to." %
+            ("Vega graph specification (JSON format)" if args.vega else "HTML document"))
+
+    if args.infile != sys.stdin:
+        # Open the specified input file.
+        args.infile = open(args.infile, "rt", encoding="UTF-8")
+    elif args.infile.isatty():
+        # No input given.  Produce an empty visualisation.
         args.infile = None
 
-    if args.outfile.isatty():
-        raise ValueError("Please specify a file name to write the %s to." %
-                         ("Vega graph specification (JSON format)" if args.vega
-                          else "HTML document"))
-
-    if args.infile is not None and args.infile != sys.stdin:
-        # Open the specified input file.
-        args.infile = open(args.infile, 'r')
-
     try:
-        create_visualisation(args.type, args.infile, args.infile2, args.outfile,
-                             args.vega, args.online, args.tidy, args.min_abs,
-                             args.min_pct_of_max, args.min_pct_of_sum,
-                             args.min_per_strand, args.bias_threshold,
-                             args.bar_width, args.padding, args.marker,
-                             args.width, args.height, args.log_scale,
-                             args.repeat_unit, args.no_alldata,
-                             args.no_aggregate, args.no_ce_length_sort,
-                             args.max_seq_len, args.jitter, args.title,
+        create_visualisation(args.type, args.infile, args.infile2, args.outfile, args.vega,
+                             args.tidy, args.min_abs, args.min_pct_of_max, args.min_pct_of_sum,
+                             args.min_per_strand, args.bias_threshold, args.bar_width,
+                             args.padding, args.marker, args.width, args.height, args.log_scale,
+                             args.repeat_unit, args.no_alldata, args.no_aggregate,
+                             args.no_ce_length_sort, args.max_seq_len, args.jitter, args.title,
                              args.allele_min_abs, args.allele_min_pct_of_max,
-                             args.allele_min_pct_of_sum,
-                             args.allele_min_correction, args.allele_min_recovery,
-                             args.allele_min_per_strand)
+                             args.allele_min_pct_of_sum, args.allele_min_correction,
+                             args.allele_min_recovery, args.allele_min_per_strand)
     except IOError as e:
         if e.errno == EPIPE:
             return
